# from PyQt5 import Qsci


# class W24GuiJsonEditor(Qsci.QsciScintilla):
#     """ Small JSON Editor with highlighter functionality
#     to make it a bit easier for the user to read the
#     JSON responses that come from the API
#     """

#     def __init__(self, parent: int = None) -> None:
#         super().__init__(parent)
#         self.setLexer(Qsci.QsciLexerJSON(self))
